package com.Smartlook.Smartlook.flutter_smartlook


import android.graphics.Rect
import android.webkit.WebView
import io.flutter.plugin.common.MethodChannel.MethodCallHandler
import com.smartlook.android.core.api.Smartlook
import com.smartlook.android.core.api.Preferences
import com.smartlook.android.core.api.enumeration.Status
import com.smartlook.android.core.api.extension.isSensitive
import com.smartlook.android.core.api.model.Properties
import com.smartlook.android.core.api.model.RecordingMask
import io.flutter.plugin.common.MethodChannel
import com.smartlook.android.core.api.model.Referrer
import com.smartlook.android.core.video.annotation.RenderingMode
import com.smartlook.android.util.logging.annotation.LogAspect
import io.flutter.plugin.common.MethodCall
import org.json.JSONArray
import org.json.JSONTokener

internal class MethodCallHandlerImpl : MethodCallHandler {
    private val smartlook: Smartlook = Smartlook.instance
    private val preferences = smartlook.preferences
    private val eventTracking: Preferences.EventTracking = preferences.eventTracking
    private val user = smartlook.user
    private val state = smartlook.state
    private var isAfterFirstStart = false
    override fun onMethodCall(call: MethodCall, result: MethodChannel.Result) {
        when (call.method) {
            "start" -> {
                smartlook.start()
                if(!isAfterFirstStart)
                {
                    preferences.isSurfaceRecordingEnabled = true
                    preferences.isAdaptiveFrameRateEnabled = false
                    preferences.renderingMode = RenderingMode.NATIVE
                    isAfterFirstStart = true
                }
            }
            "stop" -> smartlook.stop()
            "reset" -> smartlook.reset()
            "trackEvent" -> {
                val propertiesMap = call.argument<HashMap<String, String?>?>("properties")
                val properties: Properties? = propertiesFromMap(propertiesMap)
                smartlook.trackEvent((call.argument<String>("name"))!!, properties)
            }
            "trackNavigationEnter" -> {
                val propertiesMap = call.argument<HashMap<String, String?>?>("properties")
                val properties: Properties? = propertiesFromMap(propertiesMap)
                smartlook.trackNavigationEnter((call.argument<String>("name"))!!, properties)
            }
            "trackNavigationExit" -> {
                val propertiesMap = call.argument<HashMap<String, String?>?>("properties")
                val properties: Properties? = propertiesFromMap(propertiesMap)
                smartlook.trackNavigationExit((call.argument<String>("name"))!!, properties)
            }
            "setReferrer" -> {
                val referrerObject = Referrer(
                    (call.argument<String>("referrer"))!!,
                    (call.argument<String>("source"))!!
                )
                smartlook.referrer = referrerObject
            }
            "putStringEventProperty" -> {
                smartlook.eventProperties.putString(
                    (call.argument<String>("name"))!!,
                    call.argument<String>("value")
                )
            }
            "getStringEventProperty" -> {
                result.success(smartlook.eventProperties.getString((call.argument<String>("name"))!!))
                return
            }
            "removeEventProperty" -> {
                smartlook.eventProperties.remove((call.argument<String>("name"))!!)
            }
            "clearEventProperties" -> {
                smartlook.eventProperties.clear()
            }
            "setUserIdentifier" -> user.identifier = call.argument<String>("identifier")
            "setUserName" -> user.name = call.argument<String>("name")
            "setUserEmail" -> user.email = call.argument<String>("email")
            "setUserProperty" -> user.properties.putString(
                (call.argument<String>("name"))!!,
                call.argument<String>("value")
            )
            "getUserProperty" -> {
                result.success(user.properties.getString((call.argument<String>("name"))!!))
                return
            }
            "removeUserProperty" -> {
                user.properties.remove((call.argument<String>("name"))!!)
            }
            "openNew" -> {
                user.openNew()
            }
            "getUserUrl" -> {
                result.success(if (user.url == null) null else user.url!!.path)
                return
            }
            "openNewSession" -> user.session.openNew()
            "getSessionUrl" -> {
                result.success(if (user.session.url == null) null else user.session.url!!.path)
                return
            }
            "getSessionUrlWithTimeStamp" -> {
                result.success(user.session.urlWithTimestamp!!.path)
                return
            }
            "setRelayProxyHost" -> smartlook.setupConfiguration.relayProxyHost =
                call.argument<Any>("relayProxyHost") as String?
            "setProjectKey" -> preferences.projectKey = call.argument<String>("projectKey")
            "setFrameRate" -> preferences.frameRate = call.argument<Int>("frameRate")
            "getPreferencesFrameRate" -> {
                result.success(smartlook.preferences.frameRate)
                return
            }
            "setRenderingMode" -> {
                val renderingMethodKey = call.argument<Int>("renderingMode")!!
                val renderingMethod: RenderingMode =
                    if (renderingMethodKey == 0) RenderingMode.NO_RENDERING else if (renderingMethodKey == 1) RenderingMode.NATIVE else RenderingMode.WIREFRAME
                preferences.renderingMode = renderingMethod
            }
            "setRenderingModeOption" -> {
                //TODO changes for api 2.1.0
                //val renderingModeOptionKey = call.argument<Int>("renderingModeOption")!!
                //preferences.renderingModeOption = if (renderingModeOptionKey == 0) RenderingModeOption.WIREFRAME else if (renderingModeOptionKey == 1) RenderingModeOption.BLUEPRINT else if(renderingModeOptionKey ==2) RenderingModeOption.ICON_BLUEPRINT  else RenderingModeOption.SIMPLIFIED_WIREFRAME
            }
            "setJobUploadEnabled" -> {
                preferences.isUploadUsingAndroidJobsEnabled =
                    call.argument<Boolean>("uploadStatus")
            }
            "setAdaptiveFrameRateEnabled" ->
                preferences.isAdaptiveFrameRateEnabled = call.argument<Boolean>("adaptiveStatus")
            "getAdaptiveFrameRateEnabled" -> {
                result.success(state.isAdaptiveFrameRateEnabled)
                return
            }
            "setSurfaceCaptureEnabled" ->
                preferences.isSurfaceRecordingEnabled = call.argument<Boolean>("surfaceStatus")
            "getSurfaceCaptureEnabled" -> {
                result.success(state.isSurfaceRecordingEnabled)
                return
            }
            "setRecordingMask" -> {
                setRecordingMask(call.argument<String?>("maskList")) { result.success(null) }
            }
            "setWebViewEnabled" -> WebView::class.isSensitive = !call.argument<Boolean>("enabled")!!
            "eventTrackingEnableAll" -> eventTracking.enableAll()
            "eventTrackingDisableAll" -> eventTracking.disableAll()
            "eventTrackingNavigationDisableAll" -> eventTracking.navigation.disableAll()
            "setEventTrackingNavigationEnableAll" -> eventTracking.navigation.enableAll()
            "setEventTrackingNavigationActivityStatus" -> eventTracking.navigation.isActivityEnabled =
                call.argument<Boolean>("navigationActivity")!!
            "setEventTrackingNavigationFragmentStatus" -> eventTracking.navigation.isFragmentEnabled =
                call.argument<Boolean>("fragmentStatus")!!
            "setEventTrackingInteractionEnableAll" -> eventTracking.interaction.enableAll()
            "setEventTrackingInteractionDisableAll" -> eventTracking.interaction.disableAll()
            "setEventTrackingInteractionUserStatus" -> {
                val enabledUserTracking = call.argument<Boolean>("track")!!
                eventTracking.interaction.isSelectorEnabled = enabledUserTracking
                eventTracking.interaction.isTouchEnabled = enabledUserTracking
            }
            "setEventTrackingInteractionRageClickStatus" -> {
                eventTracking.interaction.isRageClickEnabled =
                    call.argument<Boolean>("rageClick")!!
            }
            "restoreDefault" -> eventTracking.default()
            "getRecordingStatus" -> {
                val status = smartlook.state.status
                result.success(if (status.isRecording) 0 else (status as Status.NotRecording).cause.ordinal + 1)
                return
            }
            "isRecording" -> {
                val status = state.status
                result.success(status.isRecording)
                return
            }
            "getProjectKey" -> {
                result.success(smartlook.state.projectKey)
                return
            }
            "getStateFrameRate" -> {
                result.success(smartlook.state.frameRate)
                return
            }
            "getRenderingMode" -> {
                result.success(smartlook.state.renderingMode.ordinal)
                return
            }
            "getRenderingModeOption" -> {
                result.success(
                    if (smartlook.state.renderingModeOption == null)
                        null
                    else smartlook.state.renderingModeOption!!.ordinal
                )
                return
            }
            "enableLogs" -> smartlook.log.allowedLogAspects = LogAspect.ALL
            else -> {
                result.notImplemented()
                return
            }
        }
        result.success(null)

    }

    private fun setRecordingMask(maskListJson:String?, resultSuccess: () -> Unit)
    {
        if (maskListJson == null)
        {
            Smartlook.instance.recordingMask = null
            resultSuccess()
            return
        }

        val jsonArray = JSONTokener(maskListJson).nextValue() as JSONArray

        val recordingMaskList:MutableList<RecordingMask.Element> = mutableListOf()

        for (i in 0 until jsonArray.length()) {
            val rect = jsonArray.getJSONObject(i).getJSONObject("rect")
            val maskType = jsonArray.getJSONObject(i).getBoolean("isCovering")
            val left = rect.getInt("left");
            val top = rect.getInt("top");
            val width = rect.getInt("width");
            val height = rect.getInt("height");
            recordingMaskList.add(RecordingMask.Element(Rect(left,top,left+width,top+height),
                if(maskType)  RecordingMask.Element.Type.COVERING
                else  RecordingMask.Element.Type.ERASING)
            );
        }
        Smartlook.instance.recordingMask = RecordingMask(recordingMaskList)
    }

    private fun propertiesFromMap(propertiesMap: HashMap<String, String?>?): Properties? {
        if (propertiesMap != null) {
            val properties = Properties()
            for ((key, value) in propertiesMap) {
                properties.putString(key, value)
            }
            return properties
        }
        return null
    }

    fun teardown() {}
}