import EventKit
import Flutter
import SmartlookAnalytics
import UIKit
import Vision
import WebKit

public class SwiftSmartlookPlugin: NSObject, FlutterPlugin {
    let smartlook = Smartlook.instance
    var isAfterFirstStart = false
    static var flutterBridge: FlutterBridge?
    
    public static func register(with registrar: FlutterPluginRegistrar) {
         let channel = FlutterMethodChannel(name: "smartlook", binaryMessenger: registrar.messenger())
         let instance = SwiftSmartlookPlugin()
         registrar.addMethodCallDelegate(instance, channel: channel)
         flutterBridge = FlutterBridge(channel: channel)

         if let flutterBridge = flutterBridge {
             Smartlook.instance.register(bridgeInterface: flutterBridge)
         }

         let eventChannel = FlutterEventChannel(name: "smartlookEvent", binaryMessenger: registrar.messenger())
         eventChannel.setStreamHandler(SwiftStreamHandler())
     }
    
    // swiftlint:disable cyclomatic_complexity
    // swiftlint:disable function_body_length
    public func handle(_ call: FlutterMethodCall, result: @escaping FlutterResult) {
        let arguments = call.arguments as? [String: Any]
        let preferences = smartlook.preferences
        var eventTracking = preferences.eventTracking
        let user = smartlook.user
        let state = smartlook.state
        
        switch call.method {
        case "start":
            smartlook.start()
            if !isAfterFirstStart {
                preferences.renderingMode = .native
                isAfterFirstStart = true
            }
            
        case "stop":
            smartlook.stop()
            
        case "reset":
            smartlook.reset()
            
        case "trackEvent":
            guard let key: String = arguments?["name"] as? String else {
                break
            }
            let properties: Properties? = properties(from: arguments?["properties"] as? [String: String?])
            
            smartlook.track(event: key, properties: properties)
            
        case "trackNavigationEnter":
            guard let key: String = arguments?["name"] as? String else {
                break
            }
            let properties: Properties? = properties(from: arguments?["properties"] as? [String: String?])
        
            smartlook.track(navigationEvent: key, direction: .enter, properties: properties)
            
        case "trackNavigationExit":
            guard let key: String = arguments?["name"] as? String else {
                break
            }
            let properties: Properties? = properties(from: arguments?["properties"] as? [String: String?])
           
            smartlook.track(navigationEvent: key, direction: .exit, properties: properties)
            
        case "putStringEventProperty":
            guard let name: String = arguments?["name"] as? String,
                  let value: String? = arguments?["value"] as? String?
            else {
                break
            }
            
            smartlook.eventProperties[name] = value
            
        case "getStringEventProperty":
            guard let name: String = arguments?["name"] as? String else {
                break
            }
            
            result(smartlook.eventProperties[name])
            return
            
        case "removeEventProperty":
            guard let name: String = arguments?["name"] as? String else {
                break
            }
            
            smartlook.eventProperties[name] = nil
            
        case "clearEventProperties":
            for key in smartlook.eventProperties.keys {
                smartlook.eventProperties[key] = nil
            }
            
        case "setUserIdentifier":
            guard let identifier: String = arguments?["identifier"] as? String else {
                break
            }
            
            user.identifier(identifier)
            
        case "setUserName":
            user.name(arguments?["name"] as? String)
            
        case "setUserEmail":
            user.email(arguments?["email"] as? String)
            
        case "setUserProperty":
            guard let name: String = arguments?["name"] as? String,
                  let value: String? = arguments?["value"] as? String?
            else {
                break
            }
            
            user.setProperty(name, to: value)
            
        case "getUserProperty":
            guard let name: String = arguments?["name"] as? String else {
                break
            }
            
            result(user[name])
            return
            
        case "removeUserPropety":
            guard let name: String = arguments?["name"] as? String else {
                break
            }
            
            user[name] = nil
            
        case "openNew":
            user.openNew()
            
        case "getUserUrl":
            result(user.url?.path)
            return
            
        case "openNewSession":
            user.session.openNew()
            
        case "getSessionUrl":
            result(user.session.url?.path)
            return
            
        case "getSessionUrlWithTimeStamp":
            result(user.session.urlWithTimestamp)
            return
            
        case "setupConfigurationRelayProxyHost":
            smartlook.setupConfiguration?.relayProxyHost(arguments?["relayProxyHost"] as? String)
            
        case "setProjectKey":
            guard let projectKey: String = arguments?["projectKey"] as? String else {
                break
            }
            
            preferences.projectKey(projectKey)
            
        case "setFrameRate":
            preferences.framerate(arguments?["frameRate"] as? Int)
            
        case "getPreferencesFrameRate":
            result(smartlook.preferences.framerate)
            return
            
        case "setRenderingMode":
            guard let renderingModeKey: Int = arguments?["renderingMode"] as? Int else {
                break
            }
            
            preferences.renderingMode(RenderingMode(int: renderingModeKey))
            
        // todo API 2.1.0
        case "setRenderingModeOption":
            break
            
        case "setAdaptiveFrameRateEnabled":
            guard let adaptiveStatus = arguments?["adaptiveStatus"] as? Bool else {
                break
            }
            
            preferences.useAdaptiveFramerate(adaptiveStatus)
            
        case "getAdaptiveFrameRateEnabled":
            result(state.usingAdaptiveFramerate)
            return
            
        case "setRecordingMask":
            guard let maskList = arguments?["maskList"] as? String? else {
                break
            }
            
            setRecordingMask(json: maskList, result: { result(nil) })
            
        case "setWebViewEnabled":
            guard let enabled = arguments?["enabled"] as? Bool else {
                break
            }
            
            smartlook.sensitivity[WKWebView.self] = !enabled
            
        case "setEventTrackingInteractionUserStatus":
            guard let track = arguments?["track"] as? Bool else {
                break
            }
            
            eventTracking?.trackUserInteraction = track
            
        case "setEventTrackingInteractionRageClickStatus":
            guard let rageClick = arguments?["rageClick"] as? Bool else {
                break
            }
            
            eventTracking?.trackRageClicks = rageClick
            
        case "eventTrackingEnableAll":
            preferences.eventTracking = EventTracking()
            
        case "eventTrackingDisableAll":
            preferences.eventTracking = EventTracking.noTracking
            
        case "restoreDefault":
            smartlook.preferences.eventTracking = EventTracking.default
            
        case "getRecordingStatus":
            let status = smartlook.state.status
            result(status.toInt())
            return
            
        case "isRecording":
            result(smartlook.state.status == .recording)
            return
            
        case "getProjectKey":
            result(smartlook.state.projectKey)
            return
            
        case "getStateFrameRate":
            result(smartlook.state.framerate)
            return
            
        case "getRenderingMode":
            let renderingMode = smartlook.state.renderingMode
            result(renderingMode.toInt())
            return
            
        // only for Android
        case "enableLogs": break
            
        case "getSurfaceCaptureEnabled":
            result(false)
            return
            
        case "setSurfaceCaptureEnabled": break
        default:
            result(FlutterMethodNotImplemented)
            return
        }
        
        result(nil)
    }

    private func setRecordingMask(json: String?, result: () -> Void) {
        guard let maskListJsonVal = json else {
            Smartlook.instance.recordingMask = nil
            result()
            return
        }

        guard let recordingMaskListData = try? JSONDecoder().decode([RecordingMaskElement].self, from: Data(maskListJsonVal.utf8)) else {
            result()
            return
        }
        
        var recordingMaskList: [MaskElement] = Array()
        for mask in recordingMaskListData {
            recordingMaskList.append(
                MaskElement(
                    rect: mask.rect.cgRect(),
                    type: mask.isCovering ? .covering : .erasing)
            )
        }
        Smartlook.instance.recordingMask = RecordingMask(elements: recordingMaskList)
    }

    private func properties(from dictionary: [String: String?]?) -> Properties? {
        guard let dictionary = dictionary else {
            return nil
        }
        
        let properties = Properties()
        for (key, value) in dictionary {
            properties[key] = value
        }
        
        return properties
    }
}

struct RecordingMaskElement: Decodable {
    let rect: Rect
    let isCovering: Bool
    enum CodingKeys: String, CodingKey {
        case rect
        case isCovering
    }

    init(from decoder: Decoder) throws {
        let rootContainer = try decoder.container(keyedBy: CodingKeys.self)
        isCovering = try rootContainer.decode(Bool.self, forKey: .isCovering)
        rect = try rootContainer.decode(Rect.self, forKey: .rect)
    }
}

extension Rect {
    func cgRect() -> CGRect {
        return CGRect(x: left, y: top, width: width , height: height)
    }
}

struct Rect: Decodable {
    let left: Int
    let top: Int
    let width: Int
    let height: Int
    enum RectCodingKeys: String, CodingKey {
        case left
        case top
        case width
        case height
    }

    init(from decoder: Decoder) throws {
        let rootContainer = try decoder.container(keyedBy: RectCodingKeys.self)
        left = try rootContainer.decode(Int.self, forKey: .left)
        top = try rootContainer.decode(Int.self, forKey: .top)
        width = try rootContainer.decode(Int.self, forKey: .width)
        height = try rootContainer.decode(Int.self, forKey: .height)
    }
}
